/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file afh.h structures for audio format handling (para_server) */

/** \cond */
#ifdef HAVE_OGGVORBIS
#define OV_AUDIO_FORMAT " ogg"
#define OV_AUDIO_FORMAT_ARRAY , "ogg"
#else
#define OV_AUDIO_FORMAT ""
#define OV_AUDIO_FORMAT_ARRAY
#endif

#ifdef HAVE_FAAD
#define AAC_AUDIO_FORMAT " aac"
#define AAC_AUDIO_FORMAT_ARRAY , "aac"
#else
#define AAC_AUDIO_FORMAT ""
#define AAC_AUDIO_FORMAT_ARRAY
#endif

#define SUPPORTED_AUDIO_FORMATS "mp3" OV_AUDIO_FORMAT AAC_AUDIO_FORMAT
#define SUPPORTED_AUDIO_FORMATS_ARRAY "mp3" OV_AUDIO_FORMAT_ARRAY \
	AAC_AUDIO_FORMAT_ARRAY, NULL

/** \endcond */

/**
 * structure for audio format handling
 *
 * There's exactly one such struct for each supported audio format. Initially,
 * only \a name and \a init are defined. During the startup process,
 * para_server calls the \a init function of each audio format handler which is
 * expected to fill in all the other function pointers.
 */
struct audio_format_handler {
	/**
	 * name of the audio format
	 */
	const char *name;
	/**
	 * typical file endings for files that can be handled by this afh.
	 */
	const char **suffixes;
	/**
	 * pointer to the audio format handler's init function
	 *
	 * Must initialize all function pointers and is assumed to succeed.
	 */
	void (*init)(struct audio_format_handler*);
	/**
	 * period of time between sending data chunks
	*/
	struct timeval chunk_tv; /* length of one chunk of data */
	/**
	 * end of file timeout - do not load new audio file until this time
	 *
	*/
	struct timeval eof_tv; /* timeout on eof */
	/**
	 * Pointer to the optional get-header function.
	 *
	 * This is called from a sender in case a new client connects in the middle of
	 * the stream.  The audio format handler may set this to NULL to indicate that
	 * this audio format does not need any special header treatment.  If non-NULL,
	 * the function it points to must return a pointer to a buffer holding the
	 * current audio file header, together with the header length.
	*/
	char *(*get_header_info)(int *header_len);
	/**
	 * check if this audio format handler can handle the file
	 *
	 * This is a  pointer to a function returning whether a given file is valid for
	 * this audio format. A negative return value indicates that this audio format
	 * handler did not recognize the given file. On success, the function is
	 * expected to return a positive value and to fill in \arg info_str, \arg
	 * chunks and \arg seconds appropriately.
	*/
	int (*get_file_info)(FILE *audio_file, char *info_str,
		long unsigned *chunks, int *seconds);
	/**
	 * cleanup function of this audio format handler
	 *
	 * This close function should deallocate any resources
	 * associated with the current audio file. In particular, it is responsible
	 * for closing the file handle. It is assumed to succeed.
	*/
	void (*close_audio_file)(void);
	/**
	 * jump to another position in the current audio file
	 *
	 * This is called if a client issued the ff or jmp command with \a request
	 * being the number of the next chunk that should be sent out. Must return a
	 * positive value on success and a negative value on errors.
	*/
	int (*reposition_stream)(long unsigned request);
	/**
	 * function responsible for reading one data chunk.
	 *
	 * \a read_chunk() must return a pointer to the next chunk of data that should
	 * be sent out, or \p NULL on errors or if the end of the file was encountered.
	 *
	 * If it returns non-NULL, \a len must contain the length of the returned
	 * buffer (which may be zero if nothing has to be sent for some reason).
	 * Otherwise, \a len is used to distinguish between the eof and the error case:
	 * It must be zero in the eof case, or negative if an error occcured.
	*/
	char * (*read_chunk)(long unsigned chunk_num, ssize_t *len);
};
